/*
 * Copyright 2022 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     https://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import org.jetbrains.kotlin.gradle.tasks.KotlinCompile

plugins {
    `kotlin-dsl`
}

group = "mahjongutils.buildlogic"
version = "0.0.1"

java {
    sourceCompatibility = JavaVersion.VERSION_11
    targetCompatibility = JavaVersion.VERSION_11
}
tasks.withType<KotlinCompile>().configureEach {
    kotlinOptions {
        jvmTarget = JavaVersion.VERSION_11.toString()
    }
}

dependencies {
    compileOnly(libs.kotlin.gradlePlugin)
    compileOnly(libs.johnrengelman.shadow.gradlePlugin)
}

gradlePlugin {
    plugins {
        register("kmplib") {
            id = "mahjongutils.buildlogic.kmplib"
            implementationClass = "mahjongutils.buildlogic.KmpLibConventionPlugin"
        }
        register("maven-publish") {
            id = "mahjongutils.buildlogic.mavenpublish"
            implementationClass = "mahjongutils.buildlogic.MavenPublishConventionPlugin"
        }
    }
}
